#! /usr/local/bin/s9 -f

; scm2html -- print Scheme code to HTML
; By Nils M Holm, 2009,2010
; See the LICENSE file of the S9fES package for terms of use
;
; Usage: scm2html [-d] [-s] [file ...]
;
; Render Scheme code contained in the given file. When no file
; is given render stdin. Write output to stdout.
;
; Options:
;
; -9  highlight S9fES (non-R4RS) procedures
; -d  write a full HTML document (default: PRE block only)
; -s  enable CSS-based syntax highlighting
; -x  highlight S9fES extension procedures
;
; The CSS2 style sheet "scheme.css" contains the default style for
; syntax and expression highlighting.
;
; NOTE: This program handles only a subset of R4RS Scheme correctly.
; Caveat utilitor!

(load-from-library "symbols.scm")
(load-from-library "read-line.scm")
(load-from-library "parse-optionsb.scm")

(define (scm2html full-html show-matches mark-s9-procs mark-s9-extns)

  (define *load-from-library* 0)

  (define LP #\()
  (define RP #\))

  (define (Prolog)
    (let ((p (if show-matches
                 '("<PRE class=scheme-hl><SPAN>")
                 '("<PRE class=scheme><SPAN>"))))
      (if full-html
          (append
           '("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\""
              "  \"http://www.w3.org/TR/html4/loose.dtd\">"
              "<HTML>"
              "<HEAD>"
              "<TITLE></TITLE>"
              "<LINK rel=\"stylesheet\" type=\"text/css\" href=\"scheme.css\">"
              "</HEAD>"
              "<BODY>")
            p)
          p)))

  (define (Epilog)
    (let ((p '("</SPAN></PRE>")))
      (if full-html
          (append p '("</BODY>" "</HTML>"))
          p)))

  (define (display* lines)
    (let* ((l*    (reverse lines))
           (last  (car l*))
           (lines (reverse (cdr l*))))
      (for-each (lambda (s)
                  (display s)
                  (newline))
                lines)
      (display last)))

  (define Color-comment    "o")
  (define Color-paren      "p")
  (define Color-symbol     "s")
  (define Color-constant   "c")
  (define Color-std-proc   "r")
  (define Color-std-syntax "y")
  (define Color-ext-proc   "x")
  (define Color-ext-syntax "z")

  (define Color #f)
  (define Bold #f)

  (define (html-display s)
    (display (apply string-append
                    (map (lambda (c)
                           (case c
                                 ((#\<) "&lt;")
                                 ((#\>) "&gt;")
                                 ((#\&) "&amp;")
                                 (else  (string c))))
                         (string->list s)))))

  (define (change-color quoted co bo thunk)
    (cond (quoted
            (thunk))
          ((and (equal? co Color) (eq? bo Bold))
            (thunk))
          (else
            (if Bold
                (begin (display "</B>")
                       (set! Bold #f)))
            (display "</SPAN><SPAN class=")
            (display co)
            (display ">")
            (if (and bo (not Bold))
                (display "<B>"))
            (set! Color co)
            (set! Bold bo)
            (thunk))))

  (define (with-color quoted co thunk)
    (change-color quoted co #f thunk))

  (define (with-bold-color quoted co thunk)
    (change-color quoted co #t thunk))

  (define (symbolic? c)
    (let ((specials "!$%&*+-./:<=>?@^_"))
      (or (char-alphabetic? c)
          (char-numeric? c)
          (and (memv c (string->list specials)) #t))))

  (define (print-paren c q)
    (cond (show-matches
            (if Bold (begin (display "</B>") (set! Bold #f)))
            (display "</SPAN>")
            (if (char=? c LP) (display "<SPAN class=n>"))
            (display "<SPAN class=")
            (display (if q Color-constant Color-paren))
            (display ">")
            (set! Color #f)
            (display c)
            (if (char=? c RP) (display "</SPAN></SPAN><SPAN>")))
          (else
            (with-color q
                        Color-paren
                        (lambda () (display c)))))
        (read-char))

  (define (r4rs-syntax? s)
    (and (memq (string->symbol s)
               (r4rs-syntax-objects))
         #t))

  (define (s9fes-syntax? s)
    (and mark-s9-procs
         (memq (string->symbol s)
               (s9fes-syntax-objects))
         #t))

  (define (r4rs-procedure? s)
    (and (memq (string->symbol s)
               (r4rs-procedures))
         #t))

  (define (s9fes-procedure? s)
    (and mark-s9-procs
         (memq (string->symbol s)
               (s9fes-procedures))
         #t))

  (define (s9fes-extension? s)
    (and mark-s9-extns
         (or (memq (string->symbol s)
                   (s9fes-extension-procedures))
             (memq (string->symbol s)
                   (s9fes-extension-symbols)))
         #t))

  (define (print-symbol-or-number c q)
    (letrec
      ((collect
         (lambda (c s)
           (cond ((symbolic? c)
                   (collect (read-char) (cons c s)))
                 (else
                   (cons c (list->string (reverse s))))))))
      (let ((c/s (collect c '())))
        (cond ((string->number (cdr c/s))
                (with-color q
                            Color-constant
                            (lambda () (html-display (cdr c/s)))))
              ((r4rs-syntax? (cdr c/s))
                (with-bold-color q
                                 Color-std-syntax
                                 (lambda () (html-display (cdr c/s)))))
              ((r4rs-procedure? (cdr c/s))
                (with-color q
                            Color-std-proc
                            (lambda () (html-display (cdr c/s)))))
              ((s9fes-syntax? (cdr c/s))
                (with-bold-color q
                                 Color-ext-syntax
                                 (lambda () (html-display (cdr c/s)))))
              ((s9fes-procedure? (cdr c/s))
                (if (string=? "load-from-library" (cdr c/s))
                    (set! *load-from-library* 2))
                (with-color q
                            Color-ext-proc
                            (lambda () (html-display (cdr c/s)))))
              ((s9fes-extension? (cdr c/s))
                (with-color q
                            Color-ext-proc
                            (lambda () (html-display (cdr c/s)))))
              (else
                (with-color q
                            Color-symbol
                            (lambda () (html-display (cdr c/s))))))
        (car c/s))))

  (define (print-string c)
    (letrec
      ((collect
         (lambda (c s esc)
           (cond ((and (char=? c #\") (not esc))
                   (list->string (reverse (cons #\" s))))
                 (else
                   (collect (read-char)
                                (cons c s)
                                (and (not esc) (char=? #\\ c))))))))
      (let* ((s  (collect c '() #t))
             (s2 (substring s 1 (- (string-length s) 1))))
        (if (= *load-from-library* 1)
            (with-color #f
                        Color-constant
                        (lambda ()
                          (display "\"<A href=\"")
                          (display s2)
                          (let ((k (string-length s2)))
                            (if (not (and (> k 3)
                                          (string=? (substring s2 (- k 4) k)
                                                    ".scm")))
                                (display ".scm")))
                          (display ".html\">")
                          (html-display s2)
                          (display "</A>\"")))
            (with-color #f
                        Color-constant
                        (lambda () (html-display s)))))
      (read-char)))

  (define (print-comment c)
    (letrec
      ((collect
         (lambda (c s)
           (cond ((char=? c #\newline)
                   (list->string (reverse s)))
                 (else
                   (collect (read-char) (cons c s)))))))
      (with-color #f
                  Color-comment
                  (lambda ()
                    (html-display (collect c '()))))
      #\newline))

  (define (print-const s)
    (with-color #f
                Color-constant
                (lambda () (html-display s)))
    (read-char))

  (define (print-name pre c)
    (letrec
      ((collect
         (lambda (c s)
           (cond ((or (char-alphabetic? c)
                      (char-numeric? c))
                   (collect (read-char) (cons c s)))
                 (else
                   (cons c (list->string (reverse s))))))))
      (with-color #f
                  Color-constant
                  (lambda ()
                    (display "#")
                    (display pre)
                    (let ((c/s (collect (read-char) (list c))))
                      (html-display (cdr c/s))
                      (car c/s))))))

  (define (print-shbang)
    (with-bold-color #f
                     Color-ext-syntax
                     (lambda ()
                       (display "#!")))
    (with-color #f
                Color-comment
                (lambda ()
                  (display (read-line))))
    #\newline)

  (define (print-hash-syntax c p)
    (let ((c (read-char)))
      (case c
            ((#\f) (cons p (print-const "#f")))
            ((#\t) (cons p (print-const "#t")))
            ((#\b) (cons p (print-name #\b (read-char))))
            ((#\d) (cons p (print-name #\d (read-char))))
            ((#\o) (cons p (print-name #\o (read-char))))
            ((#\x) (cons p (print-name #\x (read-char))))
            ((#\\) (cons p (print-name #\\ (read-char))))
            ((#\() (cons p (print-const "#(")))      ; )) balance
            ((#\!) (cons p (print-shbang)))
            (else  (error "unknown # syntax" c)))))

  (define (print-quoted c p q type)
    (with-bold-color
      q
      Color-std-syntax
      (lambda ()
        (display (if (eq? type 'quote) #\' #\`))))
      (with-color #f
                  Color-constant
                  (lambda ()
                    (print-quoted-form (read-char)
                                       p
                                       (if (eq? q 'quasiquote)
                                           'quasiquote
                                           type)))))

  (define (print-unquoted p q)
    (with-bold-color
      (eq? q 'quote)
      Color-std-syntax
      (lambda ()
        (display #\,)
        (if (char=? (peek-char) #\@)
            (display (read-char)))))
    (if (not (eq? q 'quote))
        (let ((p/c (print-quoted-form (read-char) p #f)))
          (with-color #f
                      Color-constant
                      (lambda () p/c)))
          (cons p (read-char))))

  (define (print-object c p q)
    (cond ((char=? c LP)  (cons (+ 1 p) (print-paren LP q)))
          ((char=? c RP)  (cons (- p 1) (print-paren RP q)))
          ((symbolic? c)  (cons p (print-symbol-or-number c q)))
          ((char=? c #\") (cons p (print-string c)))
          ((char=? c #\;) (cons p (print-comment c)))
          ((char=? c #\#) (print-hash-syntax c p))
          ((char=? c #\') (print-quoted c p q 'quote))
          ((char=? c #\`) (print-quoted c p q 'quasiquote))
          ((char=? c #\,) (print-unquoted p q))
          (else           (error "unknown character class" c))))

  (define (skip-whitespace c)
    (cond ((eof-object? c)
            c)
          ((char-whitespace? c)
            (display c)
            (skip-whitespace (read-char)))
          (else
            c)))

  (define (print-quoted-list c p p0 type)
    (let ((c (skip-whitespace c)))
      (cond ((eof-object? c)
              (cons p c))
            (else
              (let ((p/c (print-object c p type)))
                (if (<= (car p/c) p0)
                    p/c
                    (print-quoted-list (cdr p/c)
                                       (car p/c)
                                       p0
                                       type)))))))

  (define (print-quoted-form c p type)
    (let ((p/c (print-object c p type)))
      (if (= (car p/c) p)
          p/c
          (print-quoted-list (cdr p/c) (car p/c) p type))))

  (define (print-program c p q)
    (let ((c (skip-whitespace c)))
      (if (not (eof-object? c))
          (let ((p/c (print-object c p q)))
            (set! *load-from-library*
                  (if (zero? *load-from-library*)
                      0
                      (- *load-from-library* 1)))
            (print-program (cdr p/c) (car p/c) q)))))

   (display* (Prolog))
   (print-program (read-char) 0 #f)
   (display* (Epilog))
   (newline))

(define show-help     (option #\h #f))
(define show-matches  (option #\s #f))
(define full-html     (option #\d #f))
(define mark-s9-procs (option #\9 #f))
(define mark-s9-extns (option #\x #f))
(define options      `(,show-matches
                       ,full-html
                       ,mark-s9-procs
                       ,mark-s9-extns
                       ,show-help))

(define (usage)
  (display "Usage: scm2html [-9dsx] [file ...]")
  (newline))

(let ((files (parse-options! (sys:command-line) options usage)))
  (cond ((opt-val show-help)
          (display-usage
            `(""
              ,usage
              ""
              "Render Scheme code in HTML"
              ""
              "-9  highlight S9fES (non-R4RS) procedures"
              "-d  write full HTML document (default: PRE block only)"
              "-s  enable CSS-based syntax highlighting"
              "-x  highlight S9fES extension procedures"
              ""))
          (sys:exit 0))
        ((null? files)
          (scm2html (opt-val full-html)
                    (opt-val show-matches)
                    (opt-val mark-s9-procs)
                    (opt-val mark-s9-extns)))
        (else
          (map (lambda (file)
                 (with-input-from-file
                   file
                   (lambda ()
                     (scm2html (opt-val full-html)
                               (opt-val show-matches)
                               (opt-val mark-s9-procs)
                               (opt-val mark-s9-extns)))))
               files))))
